module configReader;

import std.regex;
import std.stdio;
// Az exceptions szabványos könyvtári modul csak
// az assertThrown és collectExceptionMsg miatt szükséges.
import std.exception;
// Az exceptions alap könyvtári modul csak az AssertError ellenőrzések miatt szükséges.
// Itt található ugyanis az AssertError osztály definíciója.
import core.exception;
import configReaderHelper;

// unit tesztek a program futtatása előtt futnak le,
// de csak akkor, ha a program -unittest kapcsolóval volt fordítva,
// különben nem fordulnak bele a programba.
unittest
{
  writeln("ConfigReader unit tesztek futtatása");

  // debug(kulcsszo)  utasításba foglalt blokk csak akkor fordul bele a programba, ha 
  // -debug=kulcsszo opcióval kerül lefordításra a progam, ahol a kulcsszo tetszőleges lehet.
  // a debug kulcsszó bárhol használható, ahova utasítást, deklarációt, vagy definíciót írhatunk,
  // Használatával még el is lehet rejteni osztálydefiníciókat.
  // Én most arra használtam, hogy ha a -debug=contract opcióval fordul a program, akkor
  // szerződéseknél (előfeltétel, utófeltétel, típus invariáns) több feltételt ellenőrzünk,
  // melyeket nem szeretnénk az átadott programban ellenőrizni, hisz ez jelentős
  // hatékonyságcsökkenéssel járna.
  debug(contract)
  {
    // Ezen kiíratás segíti a program megértését, hogy lássuk mikor futnak unit tesztek,
    // és milyen erős ellenőrzések mellett.
    writeln("Erőssebb szerződések vannak érvényben.");
  }
  writeln();
}

// Saját kivétel könnyen készíthető, az alábbi kódból csak a ConfigReaderException szót kell
// megváltoztatni, és már másolható is saját programba:
public class ConfigReaderException : Exception
{
  this(string msg, string file = __FILE__, size_t line = __LINE__)
  {
    super(msg, file, line);
  }
}

// struktúrák és az osztályok nem ugyanúgy működnek, a struktúra érték szerint adódik át,
// míg az osztályok referenciaként.
// private jelzi, hogy csak ezen modul érheti el. (Láthatóság modul szintű.)
private struct ConfigData
{
  /// Ha null, akkor nincs alapértelmezett érték, azaz kötelezően kitöltendő.
  string defaultValue;

  /// Ha null, akkor bármilyen érték megadható az adott opciónak.
  /// Nem szabad üresnek vagy egyeleműnek lennie.
  string[] allowedValues;
}

// Új osztály
// publikus, azaz más modulok is elérhetik. (Láthatóság modul szintű.)
public class ConfigReader
{
  // Alapértelmezett konstruktor készítése.
  public this()
  // Utófeltétel megadása, mivel nincs visszatérési érték, ezért arra nem adható meg feltétel.
  // Visszatérési érték szerződésben használata megtekinthető a getValue utófeltételénél.
  out
  {
    assert( m_isConfigRead == false );
  }
  // body kulcsszó jelzi, hogy vége az elő- és/vagy utófeltétel résznek, és jön a függvény törzse.
  // Akkor kell használni ha megadtunk elő- és/vagy utófeltételt.
  body
  {
    m_isConfigRead = false;
  }

  /**
   * Új opció hozzáadása.
   * param configOption Az opció neve.
   * param defaultValue Az opció alapértelmezett értéke, ha nem adjuk meg a konfig fájlban.
   *                    Ha null, akkor az opció megadása kötelező.
   *                    Ha allowedValues nem null, akkor annak egy elemével meg kell egyeznie.
   * param allowedValues Az opcióhoz adható értékek.
   *                     Ha null, akkor bármilyen érték megadható az opcióhoz.
   *                     Nem lehet üres vagy egyelemű tömb.
   */
  public void addOption(string configOption, string defaultValue, string[] allowedValues)
  // előfeltétel megadása.
  in
  {
    bool optionNameContainsWhiteSpace = !std.regex.matchFirst(configOption, `\s`).empty();
    // assert segítségével tudunk megadni ellenőrzendő feltételt, melynek teljesülnie kell.
    assert( ! optionNameContainsWhiteSpace );
    assert( ! m_isConfigRead );
    // Itt a keys.contains tulajdonság (property) lekérése valójában
    // egy contains(keys) hívás. Ez csak akkor működik, ha az első paraméter
    // (most keys) egy tömb. A contains a configReaderHelper modulban található.
    assert( ! m_configSettings.keys.contains(configOption) );
    assert( null == allowedValues || allowedValues.length >= 2 );
    assert( null == allowedValues || defaultValue == null ||  allowedValues.contains(defaultValue) );
  }
  out
  {
    assert( ! m_isConfigRead );
  }
  body
  {
    m_configSettings[configOption] = ConfigData( defaultValue, allowedValues.dup );
  }

  unittest
  {
    // Megjegyzés:
    //   Pozitív tesztnek azt nevezzük, amikor a jó működést vizsgáljuk.
    //   Negatív tesztnek pedig azt, amikor szándékosan rosszul használjuk a tesztelendő elemet,
    //   és megnézzük, hogy minden az elvárt módon történik-e.
    writeln("addOption pozitív tesztek");
    ConfigData[string] expectedConfigSettings;
    expectedConfigSettings["elhagyható_mező"] = ConfigData("alapértelmezett érték", null);
    expectedConfigSettings["kotelező_mező"] = ConfigData(null, null);
    expectedConfigSettings["kotelező_igen_vagy_nem_mező"] = ConfigData(null, ["igen","nem"]);
    expectedConfigSettings["elhagyható_igen_vagy_nem_mező"] = ConfigData("igen", ["igen","nem"]);

    ConfigReader configReader = new ConfigReader();
    configReader.addOption("elhagyható_mező", "alapértelmezett érték", null);
    configReader.addOption("kotelező_mező", null, null);
    configReader.addOption("kotelező_igen_vagy_nem_mező", null, ["igen","nem"]);
    configReader.addOption("elhagyható_igen_vagy_nem_mező", "igen", ["igen","nem"]);

    assert(configReader.m_configSettings == expectedConfigSettings);

    writeln("Siker");
    writeln();
  }

  unittest
  {
    writeln("addOption szóközt tartalmazó opcióval tesztek");
    ConfigReader configReader = new ConfigReader();
    // A következő sor az előfeltétel ellenőrzését teszteli, AssertError kivétel dobódik,
    // ha nem teljesül egy assert.
    assertThrown!AssertError( configReader.addOption("szokozos mezo", null, null) );
    writeln("Siker");
    writeln();
  }

  unittest
  {
    writeln("addOption negatív tesztek megengedett értékekre");
    ConfigReader configReader = new ConfigReader();
    configReader.addOption("confignév_null", null, null);
    assertThrown!AssertError( configReader.addOption("confignév1üres", null, [""]) );
    assertThrown!AssertError( configReader.addOption("confignév1_nem_üres", null, ["érték"]) );
    writeln("Siker");
    writeln();
  }

  unittest
  {
    writeln("addOption negatív teszt default értékekre");
    ConfigReader configReader = new ConfigReader();
    assertThrown!AssertError( configReader.addOption("confignév0", "érték", ["érték1", "érték2"] ) );
    assertThrown!AssertError( configReader.addOption("confignév0", "érték1 ", ["érték1", "érték2"] ) );
    assertThrown!AssertError( configReader.addOption("confignév0", " érték1", ["érték1", "érték2"] ) );
    assertThrown!AssertError( configReader.addOption("confignév0", "érték11", ["érték1", "érték2"] ) );
    writeln("Siker");
    writeln();
  }

  unittest
  {
    writeln("addOption mély másolás teszt");
    ConfigReader configReader = new ConfigReader();
    string configOption = "konfig_mező";
    string defaultValue = "default";
    string[] allowedValues = ["default", "érték2", "érték3"];
    configReader.addOption(configOption, defaultValue, allowedValues);
    allowedValues.length = 1; // átméretezés
    allowedValues[0] = "érték0";
    assert( 3 == configReader.m_configSettings[configOption].allowedValues.length );
    assert( "default" == configReader.m_configSettings[configOption].allowedValues[0] );
    writeln("Siker");
    writeln();
  }

  
  /**
   * Előzőleg addOption metódussal hozzáadott opciók kiolvasása a megadott konfig fájlból.
   * param configFilePath A konfigurációs fájl útvonala.
   */
  public void fillConfigValesFromFile(string configFilePath)
  in
  {
    assert( ! m_isConfigRead );
    assert(m_configValues.length == 0);
  }
  out
  {
    assert( m_isConfigRead );
  }
  body
  {
    string[] configFileLines = [];
    auto file = File(configFilePath);

    foreach (str; file.byLine) {
      configFileLines ~= str.idup;
    }
    file.close();

    fillConfigValesFromLines(configFileLines);
  }

  // Annak ellenére, hogy a setDefaultValues privát, ezen modulban (modul egy fordítási egység)
  // használható. (Láthatóság modul szintű.)
  // Ha osztály szintű láthatóságot szeretnénk, mindössze annyit kell tennünk, hogy minden osztályt
  // külön modulba (külön fájlba) helyezünk.
  /**
   * Előzőleg addOption metódussal hozzáadott opciók alapértelmezett értékeinek beállítása.
   * Ahol nem volt megadva alapértelmezett érték, ott nem lesz beállítva érték.
   */
  private void setDefaultValues()
  in
  {
    assert( ! m_isConfigRead );
    assert( m_configValues.length == 0 );
    assert( m_configSettings.length != 0 );
  }
  out
  {
    assert( ! m_isConfigRead );
    debug(contract)
    {
      assert( m_configValues.length == countConfigWithDefaultValue());
    }
  }
  body
  {
    // foreach szintaxis.
    foreach(string option; m_configSettings.keys)
    {
      string defaultValue = m_configSettings[option].defaultValue;
      if(defaultValue != null)
      {
        m_configValues[option] = defaultValue;
      }
    }
  }

  unittest
  {
    writeln("setDefaultValues teszt");
    ConfigReader configReader = new ConfigReader();
    configReader.addOption("kitöltendő1", "érték1", ["érték1", "érték2"]);
    configReader.addOption("nem_kitöltendő1", null, ["érték3", "érték4"]);
    configReader.addOption("kitöltendő2", "érték5", null);
    configReader.addOption("nem_kitöltendő2", null, null);
    configReader.addOption("kitöltendő3", "érték6", null);
    configReader.addOption("kitöltendő4", "érték7", null);
    configReader.setDefaultValues();
    assert( "érték1" == configReader.m_configValues["kitöltendő1"] );
    assert( "érték5" == configReader.m_configValues["kitöltendő2"] );
    assert( "érték6" == configReader.m_configValues["kitöltendő3"] );
    assert( "érték7" == configReader.m_configValues["kitöltendő4"] );
    assert( 4 == configReader.m_configValues.length );
    writeln("Siker");
    writeln();
  }

  /**
   * Az addOption metódussal hozzáadott alapételmezett értékekkel rendelkezó opciók számát
   * adja vissza.
   */
  private int countConfigWithDefaultValue()
  {
    int sum = 0;
    foreach(string option; m_configSettings.keys)
    {
      if( null != m_configSettings[option].defaultValue )
      {
        ++sum;
      }
    }
    return sum;
  }

  unittest
  {
    writeln("countConfigWithDefaultValue teszt");
    ConfigReader configReader = new ConfigReader();
    assert( 0 == configReader.countConfigWithDefaultValue() );
    configReader.addOption("kitöltendő1", null, ["érték1", "érték2"]);
    assert( 0 == configReader.countConfigWithDefaultValue() );
    configReader.addOption("nem_kitöltendő1", "érték3", ["érték3", "érték4"]);
    assert( 1 == configReader.countConfigWithDefaultValue() );
    configReader.addOption("kitöltendő2", null, null);
    assert( 1 == configReader.countConfigWithDefaultValue() );
    configReader.addOption("nem_kitöltendő2", "érték5", null);
    assert( 2 == configReader.countConfigWithDefaultValue() );
    configReader.addOption("kitöltendő3", null, null);
    assert( 2 == configReader.countConfigWithDefaultValue() );
    configReader.addOption("kitöltendő4", null, null);
    assert( 2 == configReader.countConfigWithDefaultValue() );
    configReader.addOption("nem_kitöltendő3", "érték6", null);
    assert( 3 == configReader.countConfigWithDefaultValue() );
    configReader.addOption("nem_kitöltendő4", "érték7", null);
    assert( 4 == configReader.countConfigWithDefaultValue() );
    assert( 4 == configReader.countConfigWithDefaultValue() );
    writeln("Siker");
    writeln();
  }

  /**
   * Előzőleg addOption metódussal hozzáadott opciók kiolvasása a megadott konfig sorokból
   * param configFileLines A konfigurációs fájl sorainak tartalma.
   */
  private void fillConfigValesFromLines(const string[] configFileLines)
  in
  {
    assert( ! m_isConfigRead );
    assert( m_configValues.length == 0 );
    assert( m_configSettings.length != 0 );
  }
  out
  {
    assert( m_isConfigRead );
    assert( m_configValues.length == m_configSettings.length);
  }
  body
  {
    setDefaultValues();
    foreach(string line; configFileLines)
    {
      if( auto match = std.regex.matchFirst(line, `^\s*$`) )
      {
        assert( match.pre == "" );
        assert( match.post == "" );
        continue;
      }

      if( auto match = std.regex.matchFirst(line, `^\s*#`) )
      {
        assert( match.pre == "" );
        continue;
      }

      if( auto match = std.regex.matchFirst(line, `^\s*(\S+)\s*=(.*)$`) )
      {
        assert( match.pre == "" );
        assert( match.post == "" );

        string option = match.captures[1];
        string value = match.captures[2];

        // Kezdő fehér szóközök törlése a value szövegből.
        value = std.regex.replaceFirst(value, std.regex.regex(`^\s*(\S.*)$`), `$1`);

        // A value szöveg végén lévő esetleges fehér szóközök törlése.
        value = std.regex.replaceFirst(value, std.regex.regex(`^(.*\S)\s*$`), `$1`);
        if ( !m_configSettings.keys.contains(option) )
        {
          throw new ConfigReaderException( option ~ " opció ismeretlen." );
        }

        if( m_configSettings[option].allowedValues != null &&
            !m_configSettings[option].allowedValues.contains(value) )
        {
          // Itt a allowedValues.toString tulajdonság (property) lekérése valójában
          // egy toString(allowedValues) hívás. Ez csak akkor működik, ha az első paraméter
          // (most allowedValues) egy tömb. A toString a configReaderHelper modulban található.
          throw new ConfigReaderException( option ~
                                           " opció értéke csak a következők egyike lehet: " ~
                                           m_configSettings[option].allowedValues.toString );
        }

        m_configValues[option] = value;
      }
      else
      {
        throw new ConfigReaderException( "Ez a sor nem felel meg a szintaxisnak: " ~ line );
      }
    }
    if( m_configValues.length != m_configSettings.length)
    {
      string[] filledConfigOptions = m_configValues.keys;
      foreach( option; m_configSettings.keys )
      {
        if( !filledConfigOptions.contains( option ) )
        {
          throw new ConfigReaderException( "Kötelező opció nincs megadva: " ~ option );
        }
      }
    }
    m_isConfigRead = true;
  }

  /**
   * Az addOption metódussal hozzáadott opciók közül a megadott opció értékét olvassa ki.
   * Hívása előtt a konfig fájlt már be kellett olvasni.
   * param configOption Azon opció, melynek aktuális értékére vagyunk kiváncsiak.
   */
  public string getValue(string configOption)
  in
  {
    assert(m_configValues.keys.contains(configOption));
    assert( m_isConfigRead );
  }
  // Utófeltétel megadása visszatérési érték hivatkozással.
  // Ha van a műveletnek visszatérési értéke, akkor opcionálisan szerepelhet zárójelben egy
  // változó, melybe el lesz tárolva a visszatérési érték,
  // hogy a szerződésekben fel lehessen használni.
  out (result)
  {
    assert(result == m_configValues[configOption]);
    assert( m_isConfigRead );
  }
  body
  {
    return m_configValues[configOption];
  }

  unittest
  {
    writeln("getValue teszt");
    ConfigReader configReader = new ConfigReader();
    configReader.addOption("mező1", null, null);
    configReader.addOption("mező2", null, null);
    configReader.addOption("mező_három", null, null);
    configReader.addOption("mezőNégy", null, null);
    configReader.addOption("asfőf", null, null);
    configReader.addOption("44", null, null);
    configReader.fillConfigValesFromLines( [
                                             "mező1 = érték1",
                                             "mező2 = asdfgh",
                                             "mező_három = 123",
                                             "mezőNégy = áéí",
                                             "asfőf = asfőf",
                                             "44 = 45"
                                           ] );
    assert( "érték1" == configReader.getValue("mező1") );
    assert( "123" == configReader.getValue("mező_három") );
    assert( "asfőf" == configReader.getValue("asfőf") );
    assert( "áéí" == configReader.getValue("mezőNégy") );
    assert( "45" == configReader.getValue("44") );
    assert( "asdfgh" == configReader.getValue("mező2") );
    writeln("Siker");
    writeln();
  }


  private bool               m_isConfigRead;
  private string[string]     m_configValues;
  private ConfigData[string] m_configSettings;

  // A ConfigReader típusinvariánsa.
  // Konstruktor után, destruktor előtt, és minden publikus metódus meghívása előtt és után lefut.
  // Ha a publikus metódusnak van elő- és/vagy utófeltétele is, akkor az előfeltétel után,
  // és az utófeltétel előtt fut le.
  // Privát metódusok előtt és után nem fut le az ellenőrzés, hiszen ezeket kívülről nem hívják,
  // és el is szabad rontaniuk az invariánst, hiszen többnyire publikus műveletek hívják őket,
  // azonban az a szabály, hogy a következő publikus metódus hívása előtt helyre kell hozni az
  // invariánst.
  // Ez azért nagyon hasznos, mert a publikus metódusok sok kissebb privát metódusra bonthatóak,
  // melyek eggyütt megtartják az invariánst, de köztes állapotban az nem áll fenn.
  invariant()
  {
    assert(!m_isConfigRead || m_configSettings.length == m_configValues.length);

    // A debug(contract) itt kifejezetten fontos, hiszen ebben a blockban megadott részek futási
    // ideje lineáris, nem pedig konstans, továbbá minden publikus metódus meghívásakor kétszer
    // is lefutnak.
    debug(contract)
    {
      // Függvénybe ágyazott függvények.
      // Hasznosak, ha csak egy függvényben szeretnénk használni őket.
      // Megjegyzés: Ugyanígy beágyazható függvénybe akár egy új osztály definíciója is.
      void assertConfigSettingsIn(const ref ConfigData configData)
      {
        if(configData.allowedValues != null)
        {
          assert(configData.allowedValues.length >= 2);
          assert(null == configData.defaultValue ||
                 configData.allowedValues.contains(configData.defaultValue));
        }
      }
      void assertConfigValueFor(string option)
      {
        if( ! m_isConfigRead )
        {
          return;
        }
        // allowedValues egy alias (referencia) a m_configSettings[option].allowedValues tömbre.
        // azaz nem történik másolás.
        const string[] allowedValues = m_configSettings[option].allowedValues;

        assert(m_configValues[option] != null);
        if( allowedValues != null )
        {
          assert(allowedValues.contains(m_configValues[option]));
        }
      }

      if( m_isConfigRead )
      {
        assert( m_configSettings.keys == m_configValues.keys );
      }

      foreach(const string option; m_configSettings.keys)
      {
        assertConfigSettingsIn(m_configSettings[option]);
        assertConfigValueFor(option);
      }
    }

  }

}

unittest
{
  writeln("Egyszerű egysoros konfig fájl teszt");
  ConfigReader configReader = new ConfigReader();
  string configOption = "konfig_mező";
  string configValue  = "érték1";
  configReader.addOption(configOption, "érték0", null);
  configReader.fillConfigValesFromLines([configOption ~ " = " ~ configValue]);
  assert( configReader.getValue(configOption) == configValue );
  writeln("Siker");
  writeln();
}
  
unittest
{
  writeln("ConfigReader funkcionális teszt");
  // Kivétel dobásának szükségességét az exception szabványos csomagban lévő
  // assertThrown vagy collectExceptionMsg template segítségével tehetjük meg.
  // felkiáltójel után a szükséges kivétel típusát (template paraméter) kell megadni,
  // zárójel után pedig a kifejezést.
  // Az assertThrown azt ellenőrzi, hogy kobódott-e kivétel, illetve az a kivétel a
  // megadott kivétel-e.  Ha más kivétel dobódik, vagy nem dobódik kivétel,
  // akkor sikertelen lesz a teszt, AssertError kivétel dobódik.
  // Ezzel szemben a collectExceptionMsg mindössze vár egy kivételt, és visszatér a kivételbe
  // állított üzenettel.
  // Példa kivétel ellenőrzésére:
  //   assertThrown!SzuksegesKivetel( configReader.fillConfigValesFromLines(configFileLines) );
  // Példa kivétel és annak szövegének ellenőrzésére:
  //   assert( "Szükséges üzenet" ==
  //           collectExceptionMsg!SzuksegesKivetel(
  //             configReader.fillConfigValesFromLines(configFileLines) ) );

  // createConfigReader csak ezen teszten belül definiált, ha nem fordul bele a unit teszt a kódba,
  // azaz ha nincs megadva a -unittest, akkor ezen fügvény sem fordul bele a kódba fölöslegesen.
  ConfigReader createConfigReader()
  {
    ConfigReader configReader = new ConfigReader();
    configReader.addOption("readingmode", "readonly", ["readonly", "readwrite"]);
    configReader.addOption("admin", null, ["yes", "no"]);
    configReader.addOption("mynickname", "Anonymus", null);
    configReader.addOption("myusername", null, null);
    return configReader;
  }

  string[] configFileLines = null;
  ConfigReader configReader = null;

  // Amikor több szóköz is szerepel
  configReader = createConfigReader();
  configFileLines =
    [
      "# aligned",
      "readingmode   = readonly",
      "admin         = no",
      "mynickname    = Jó Béla",
      "myusername    = jobela"
    ];
  configReader.fillConfigValesFromLines(configFileLines);
  assert( "readonly" == configReader.getValue("readingmode") );
  assert( "no" == configReader.getValue("admin") );
  assert( "Jó Béla" == configReader.getValue("mynickname") );
  assert( "jobela" == configReader.getValue("myusername") );
  // másodszor is lekéhető.
  assert( "Jó Béla" == configReader.getValue("mynickname") );
  assertThrown!AssertError( configReader.getValue("nincsilyen") );
  assertThrown!AssertError( configReader.getValue("admin2") );

  // Amikor megjegyzések is szerepelnek.
  configReader = createConfigReader();
  configFileLines =
    [
      "# readingmode = readwrite",
      "readingmode = readonly",
      "# readingmode = readwrite",
      "   #admin = yes",
      "admin = no",
      "   #   admin = yes",
      "   # admin = yes",
      "mynickname = Jó Béla",
      "myusername = jobela"
    ];
  configReader.fillConfigValesFromLines(configFileLines);
  assert( "readonly" == configReader.getValue("readingmode") );
  assert( "no" == configReader.getValue("admin") );
  assert( "Jó Béla" == configReader.getValue("mynickname") );
  assert( "jobela" == configReader.getValue("myusername") );
  // másodszor is lekéhető.
  assert( "Jó Béla" == configReader.getValue("mynickname") );
  assertThrown!AssertError( configReader.getValue("nincsilyen") );
  assertThrown!AssertError( configReader.getValue("admin2") );

  // Amikor minden meg van adva.
  configReader = createConfigReader();
  configFileLines =
    [
      "readingmode = readonly",
      "admin = no",
      "mynickname = Jó Béla",
      "myusername = jobela"
    ];
  configReader.fillConfigValesFromLines(configFileLines);
  assert( "readonly" == configReader.getValue("readingmode") );
  assert( "no" == configReader.getValue("admin") );
  assert( "Jó Béla" == configReader.getValue("mynickname") );
  assert( "jobela" == configReader.getValue("myusername") );
  // másodszor is lekéhető.
  assert( "Jó Béla" == configReader.getValue("mynickname") );
  assertThrown!AssertError( configReader.getValue("nincsilyen") );
  assertThrown!AssertError( configReader.getValue("admin2") );

  // Amikor egyvalami hiányzik
  configReader = createConfigReader();
  configFileLines =
    [
      "readingmode = readwrite",
      "admin = yes",
      "myusername = jobela"
    ];
  configReader.fillConfigValesFromLines(configFileLines);
  assert( "readwrite" == configReader.getValue("readingmode") );
  assert( "yes" == configReader.getValue("admin") );
  assert( "Anonymus" == configReader.getValue("mynickname") );
  assert( "jobela" == configReader.getValue("myusername") );
  // Másodszor is lekérhető.
  assert( "Anonymus" == configReader.getValue("mynickname") );
  assertThrown!AssertError( configReader.getValue("nincsilyen") );
  assertThrown!AssertError( configReader.getValue("admin2") );

  // Amikor csak a kötelező mezők vannak megadva.
  configReader = createConfigReader();
  configFileLines =
    [
      "admin = no",
      "myusername = jobela"
    ];
  configReader.fillConfigValesFromLines(configFileLines);
  assert( "readonly" == configReader.getValue("readingmode") );
  assert( "no" == configReader.getValue("admin") );
  assert( "Anonymus" == configReader.getValue("mynickname") );
  assert( "jobela" == configReader.getValue("myusername") );
  // másodszor is lekéhető.
  assert( "Anonymus" == configReader.getValue("mynickname") );
  assertThrown!AssertError( configReader.getValue("nincsilyen") );
  assertThrown!AssertError( configReader.getValue("admin2") );

  // Ismeretlen opció van jelen.
  configReader = createConfigReader();
  configFileLines =
    [
      "readingmode = readonly",
      "admin = no",
      "mynickname = Jó Béla",
      "myusername = jobela",
      "nosuch = no"
    ];
  assert( "nosuch opció ismeretlen." ==
          collectExceptionMsg!ConfigReaderException(
            configReader.fillConfigValesFromLines(configFileLines) ) );

  // readingmode nem megfelelő értékkel szerepel.
  configReader = createConfigReader();
  configFileLines =
    [
      "readingmode = readonl",
      "admin = no",
      "mynickname = Jó Béla",
      "myusername = jobela"
    ];
  assert( "readingmode opció értéke csak a következők egyike lehet: readonly, readwrite" == 
          collectExceptionMsg!ConfigReaderException(
            configReader.fillConfigValesFromLines(configFileLines) ) );

  // admin nem megfelelő értékkel szerepel.
  configReader = createConfigReader();
  configFileLines =
    [
      "readingmode = readonly",
      "admin = n",
      "mynickname = Jó Béla",
      "myusername = jobela"
    ];
  assert( "admin opció értéke csak a következők egyike lehet: yes, no" ==
          collectExceptionMsg!ConfigReaderException(
            configReader.fillConfigValesFromLines(configFileLines) ) );

  // Rossz szintaxis szerepel a fájlban.
  configReader = createConfigReader();
  configFileLines =
    [
      "readingmode: readonly",
      "admin = n",
      "mynickname = Jó Béla",
      "myusername = jobela"
    ];
  assert( "Ez a sor nem felel meg a szintaxisnak: readingmode: readonly" ==
          collectExceptionMsg!ConfigReaderException(
            configReader.fillConfigValesFromLines(configFileLines) ) );

  // Kötelező opció hiányzik: admin
  configReader = createConfigReader();
  configFileLines =
    [
      "readingmode = readonly",
      "mynickname = Jó Béla",
      "myusername = jobela"
    ];
  assert( "Kötelező opció nincs megadva: admin" ==
          collectExceptionMsg!ConfigReaderException(
            configReader.fillConfigValesFromLines(configFileLines) ) );

  // Kötelező opció hiányzik: myusername
  configReader = createConfigReader();
  configFileLines =
    [
      "readingmode = readonly",
      "admin = no",
      "mynickname = Jó Béla",
    ];
  assert( "Kötelező opció nincs megadva: myusername" ==
          collectExceptionMsg!ConfigReaderException(
            configReader.fillConfigValesFromLines(configFileLines) ) );

  // Összes kötelező opció hiányzik.
  configReader = createConfigReader();
  configFileLines =
    [
      "readingmode = readonly",
      "mynickname = Jó Béla",
    ];
  assertThrown!ConfigReaderException( configReader.fillConfigValesFromLines(configFileLines) );

  // Összes kötelező opció hiányzik, mivel nincs egy sor sem.
  configReader = createConfigReader();
  configFileLines =
    [
    ];
  assertThrown!ConfigReaderException( configReader.fillConfigValesFromLines(configFileLines) );

  writeln("Siker");
  writeln();
}

unittest
{
  writeln("ConfigReader unit tesztek sikeresen lefutottak.");
  writeln();
  writeln();
  writeln();
  writeln();
}


